#ifndef ICLIENT_HH
#define ICLIENT_HH

#include <sstream>

#include <errno.h>
#include <netinet/tcp.h>
#include <netdb.h> 
#include <unistd.h>
#include <fcntl.h>

#include <imsg.hh>
#include <message.hh>
#include <msgConstants.hh>

class iclient {

private:
	int sock;

public:
	enum received{STAT_ERR, STAT_EMPT, STAT_RECV};

	iclient(){}
	
	~iclient(){
		disconnect();
	}

	bool connect(const char *serverName, int portNumber){
		//Set up the socket
		struct addrinfo * address, *current;
		struct addrinfo hints;
		memset(&hints, 0, sizeof(struct addrinfo)); //Create empty hint
		hints.ai_family=AF_UNSPEC; //For both IPv4 and IPv6
		hints.ai_socktype=SOCK_STREAM;
		std::stringstream port;
		port<<portNumber;
		int result=getaddrinfo("localhost", port.str().c_str(), &hints, &address);
		if(result!=0){
			if(result==EAI_SYSTEM){
				int errnum=errno;
				std::cerr<<"Error while creating the address :"<<std::endl;
				std::cerr<<strerror(errnum)<<std::endl;
				return false;
			}else{
				std::cerr<<"Error while creating the address :"<<std::endl;
				std::cerr<<gai_strerror(result)<<std::endl;
				return false;
			}
		}
		
		for (current=address; current != NULL; current=current->ai_next) {
			sock=socket(current->ai_family, current->ai_socktype, current->ai_protocol);
			if(sock==-1){
				continue;
			}
			
			//Allow multiple connexions
			int reuse_addr=1;
			if(setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &reuse_addr, sizeof(reuse_addr))!=0){
				close(sock);
				continue;
			}
			
			if(::connect(sock, current->ai_addr, current->ai_addrlen) == 0)
				break;
			
			close(sock);
		}
		
		freeaddrinfo(address);
		
		if(current==NULL || sock==-1 || result==-1){
			int errnum=errno;
			std::cerr<<"Error while creating socket :"<<std::endl;
			std::cerr<<strerror(errnum)<<std::endl;
			return false;
		}
		
		//Make the socket non blocking
		int options;
		options=fcntl(sock, F_GETFL);
		if(options<0){
			std::cerr<<"Problem while getting previous options from the master socket"<<std::endl;
			close(sock);
			return false;
		}
		options=(options | O_NONBLOCK);
		if(fcntl(sock, F_SETFL, options)<0){
			std::cerr<<"Could not set master socket's properties"<<std::endl;
			close(sock);
			return false;
		}

		return true;
	}

	void subscribe(const char *name){
		message post;
		post.encode(DEFAULT_SERVER_NAME, name);
		send(post);
	}

	void disconnect(){
		close(sock);		
	}

	bool checkReadSocket(bool blocking){
		bool quit=!blocking;
		ssize_t read;
		
		do{
			errno=0;
			read=::recv(sock, NULL, 0, 0);
			if(read<0 && errno!=0 && errno!=EAGAIN && errno!=EWOULDBLOCK){
				std::cerr<<"Error on the socket while reading :"<<std::endl;
				std::cerr<<strerror(errno)<<std::endl;
				return false;
			}
			if(read==0) quit=true;
			else usleep(10);
		}while(quit!=true);
		
		return(read==0);
	}

	bool send(message& data){
		imsg msg;
		msg.encode(data.data());
		ssize_t n;
		size_t tot = 0;
		size_t bleft = msg.length();
		while(tot < msg.length()) {
			n = ::send(sock, msg.data() + tot, bleft, MSG_NOSIGNAL);
			if(n < 1) break;
			tot += n;
			bleft -= tot;
		}
		return (tot == msg.length());
	}

	std::pair<received, message> receive(){
		message result;
		imsg msg;
		ssize_t res;
		char header[msg.hsize()];
		errno=0;
		res = ::recv(sock, header, msg.hsize(), 0);
		if(res==0){ //Socket error
			return std::make_pair(STAT_ERR, result);
		}else if(res<0){ //Nothing to read
			return std::make_pair(STAT_EMPT, result);		
		}
		if(res != static_cast<ssize_t>(msg.hsize())){
			return std::make_pair(STAT_ERR, result);
		}
		msg.decodeHeader(header);
		char body[msg.blength()];
		res = ::recv(sock, body, msg.blength(), MSG_WAITALL);
		if(res != static_cast<ssize_t>(msg.blength())){
			return std::make_pair(STAT_ERR, result);
		}
		msg.decodeBody(body);
		result.decode(msg.body());
		return std::make_pair(STAT_RECV, result);
	}
};

#endif
